#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <fcntl.h>
#include <malloc.h>
#include <signal.h>
#include <unistd.h>

#include <glib.h>
#include <argp.h>
#include "vm2.h"
#include "channels.h"
#include "cells_sh.h"
#include <SDL/SDL.h>       

#define WHEN 100

Machine vm;
int cap1,cap2,cap3;

int issdl;
SDL_Event event;         

const char *argp_program_version = PACKAGE "-" VERSION;
const char *argp_program_bug_address = "mar <marbug@1010.co.uk>";
static char doc[] = "cellular p2p VM";
static char args_doc[] = "[FILE] ...";
static struct argp_option options[] = {
  {"memsz", 'm', "SIZE", 0, "memory size"},
   {"input", 'i', 0, 0, "O_RDONLY files follow (default)"},
  {"output", 'o', 0, 0, "O_WRONLY files follow"},
  {"inout", 'y', 0, 0, "O_RDWR files follow"},
  {"video", 'v', "VIDEODEV", 0, "video capabilities + videodev"},
  {"net", 'n', 0, 0, "netwk capabilities"},
  {"filesystem", 'f', "STARTFILE", 0, "filesystem capabilities"},
  {"sdlout",'s',0,0, "sdl out"},
  {"vectorout",'l',0,0, "vector out"},
  {"sniffer",'t',0,0, "sniffer eth0 in"},
  {"http",'w',0,0, "webpage out"},
  {"midiin",'q',"MIDIDEV",0, "midi input"},
  {"loopfile",'p',"LOOPFILE",0,"loopfile specified"},
  {"sharedmem",'x',0,0,"sharedmemory"},
  {"broadcast",'b',0,0,"broadcast"},
  {0}
};

struct Options
{
  unsigned int memsz;
  int delay;
};
struct Options opt;

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct Options *opt = state->input;
  static int files_flags = O_RDONLY;
  Channel *ch;
  _mirror *mir;

	mir=vm.mirror;
 
mir->soutcount=0;
	mir->voutcount=0;

 issdl=0;
 


  switch (key) {
  case 'm':
    opt->memsz = atoi (arg);
    break;
  case 'd':
    opt->delay = atoi (arg);
    break;
  case 'i':
    files_flags = O_RDONLY;
    break;
  case 'o':
    files_flags = O_WRONLY;
    break;
  case 'y':
    files_flags = O_RDWR;
    break;
  case 'v':
    if (ch = videoch (arg, 320, 240))
{      vm.channels = g_list_append (vm.channels, ch);


}
    else
      g_warning ("v1: Can't open video");
    break;

  case 'q':
  if (ch=midich(arg))
  	vm.channels = g_list_append (vm.channels, ch);
  else
   g_warning ("v1: Can't open midi");
     break;
    
  case 'x':
  if (ch=sharedch())
  	vm.channels = g_list_append (vm.channels, ch);
  else
   g_warning ("s1: Can't open shared");
     break;
    
    case 's':
    if (ch = sdlch(1024,768,8,0)){
      vm.channels = g_list_append (vm.channels, ch);
      issdl=1;


}
   
    break;

	case 'l':
	 if (ch = sdlch(1024,768,4,1)){
	       vm.channels = g_list_append (vm.channels, ch);
	             issdl=1;
	                 }
	                     break;      

  case 'n':
    if (ch = netch ()) {
      vm.channels = g_list_append (vm.channels, ch);
    }
    break;

  case 'b':
    if (ch = bcastch ()) {
      vm.channels = g_list_append (vm.channels, ch);
    }
    break;


  case 'p':
    if (ch = loopch (arg)) {
      vm.channels = g_list_append (vm.channels, ch);
    }
    break;
    
     case 't':
         if (ch = sniffch ()) {
         vm.channels = g_list_append (vm.channels, ch);
                   }
                       break;             

  case 'w':
    if (ch=webch()){ 
      vm.channels = g_list_append (vm.channels, ch);
    }


    break;

    
    
  case 'f':
    if (ch = filesystemch (arg))
      vm.channels = g_list_append (vm.channels, ch);
    else
      g_warning ("f1: filesys problems");
    break;
    
  case ARGP_KEY_ARG:
    if (ch = filech (arg, files_flags)) {

      /*	g_log (0, G_LOG_LEVEL_DEBUG, "mirror");*/

      vm.channels = g_list_append (vm.channels, ch);


    }
    break;
  case ARGP_KEY_END:
    vm.chcnt = g_list_length (vm.channels);
    break;
  default:
    return ARGP_ERR_UNKNOWN;
  }


  return 0;
}

static struct argp argp = { options, parse_opt, args_doc, doc };

struct BootLoader
{
  unsigned char cop;
  unsigned char ch;
  unsigned int memsz;
}
__attribute__ ((packed));
struct BootLoader program = { 8, 0 };
struct Shared *shared;
int shmid = -1, shmid2 = -1;

int
destroySh ()
{
  shmid = shmget (SHMKEY, 1, IPC_CREAT | 0666);
  shmid2 = shmget (SHMKEY2, 1, IPC_CREAT | 0666);
  if (shmid >= 0)
    shmctl (shmid, IPC_RMID, 0);
  if (shmid2 >= 0)
    shmctl (shmid2, IPC_RMID, 0);
}

int
prepareMachine ()
{
  Channel *ch;
  int fd;
  destroySh ();

  if ((shmid = shmget (SHMKEY, sizeof (struct Shared), IPC_CREAT | 0666)) < 0) {
    perror ("shmget(cells)");
    exit (1);
  }
  if ((shared = shmat (shmid, NULL, 0)) == 0) {
    perror ("shmat(cells)");
    exit (1);
  }

  if ((shmid2 = shmget (SHMKEY2, opt.memsz, IPC_CREAT | 0666)) < 0) {
    perror ("shmget(cells)");
    exit (1);
  }
  if ((vm.mem = shmat (shmid2, NULL, 0)) == 0) {
    perror ("shmat(cells)");
    exit (1);
  }

  vm.memsz = shared->memsz = opt.memsz;
  shared->reboot = 0;
  memset (shared->xp, 0, XP_TAIL * sizeof (int));
/*  memcpy (vm.mem, &program, sizeof (program));*/

/* instead init cellspace from /dev/urandom then parse cells in resetmachine -- reset curcell to first and run from there */
/* mark cell boundary with 42 */

  /*fd=open("/dev/urandom", O_RDONLY);*/
    ch=g_list_nth_data (vm.channels,0);

    ch->rd (ch,vm.mem,vm.memsz);


  


  resetMachine (&vm);
}

void
cleanup (void)
{
  void dstr (gpointer data, gpointer udata)
  {
    Channel *ch = (Channel *) data;
    g_assert (ch->destroy);
    ch->destroy (ch);
  }
  g_list_foreach (vm.channels, dstr, 0);

  if (shmid >= 0)
    shmctl (shmid, IPC_RMID, 0);
  if (shmid2 >= 0)
    shmctl (shmid2, IPC_RMID, 0);
}

volatile sig_atomic_t sig_term_in_progress = 0;
void
sig_term (int sig)
{
  if (sig_term_in_progress)
    raise (sig);
  sig_term_in_progress = 1;

  cleanup ();
  signal (sig, SIG_DFL);
  raise (sig);
}

int
main (int argc, char *argv[])
{
  GList *pcur;
  int i, semid,fd, x, bpp;
  int fnoc=1,cnt=1,local=0,local1=0;
  Channel *ch;
  SDLchannel *sd;
  struct M_filesystem *fc;
/*  struct sembuf semtakeover[2] =
    { {0, 0, SEM_UNDO}, {0, 1, SEM_UNDO | IPC_NOWAIT} }, semgiveup[1] =
  { {
  0, -1, SEM_UNDO | IPC_NOWAIT}
  };

  if ((semid = semget (SEMKEY, 1, IPC_CREAT | 0666)) < 0) {
    perror ("semget");
    exit (1);
  }*/

  initMachine (&vm);
  cap1=0;
  cap2=0;
  cap3=0;

  opt.memsz = 1000;
  opt.delay = 0;

		vm.mirror=malloc(sizeof ( _mirror));
		vm.mirror->vid=0;

  argp_parse(&argp,argc,argv,ARGP_IN_ORDER,0,&opt);

		
  g_log (0, G_LOG_LEVEL_DEBUG, "HERE");


  /* sort mirrorfns depends on channels -- go thru all channels */


  vm.mirror->vid=0;
  vm.mirror->snd=0;
  vm.mirror->filevid-0;
  vm.mirror->filesnd=0;

  for (x=0;x<g_list_length (vm.channels);x++){
    ch=g_list_nth_data (vm.channels,x);
    if (ch->mirrorname==1) {
vm.mirror->snd=ch->mirrorrw;
vm.mirror->data1=ch->data;
vm.mirror->sndchan= x;
 cap1=1;
    }
    if (ch->mirrorname==2 && ch->mirrorrw==1){
      vm.mirror->vid=vm.mirror->vid|1;
      vm.mirror->data2= (VideoChannel *) ch->data;
    }
    if (ch->mirrorname==2 && ch->mirrorrw==2){
      vm.mirror->vid=vm.mirror->vid|2;
      vm.mirror->vidchan=x;
      cap2=1;
    }
    if (ch->mirrorname==3) {
      cap3=1;      
local=1;
    vm.mirror->filechan=x;
    vm.mirror->data3=(struct M_filesystem *)ch->data;    

}

    if (ch->mirrorname==4) local1=1;

  }
  if (local==1 && vm.mirror->vid>=2) 
vm.mirror->filevid=1;
  if (local==1 && vm.mirror->snd>=2)
vm.mirror->filesnd=1;

  vm.mirror->filesndcount=0;
  vm.mirror->filevidcount=0;


  program.memsz = opt.memsz;
  atexit (cleanup);

  prepareMachine ();

/* seed */
srandom(time(0));

/* do all sdl setup here */

 g_log (0, G_LOG_LEVEL_DEBUG, "cap1,2,3%d  %d  %d",cap1,cap2,cap3);

 if (cap1==1 && cap2==0 && cap3==1){
   
if (SDL_Init (SDL_INIT_AUDIO) < 0) {
    fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());         
    exit (1);                                                                   
  }                                                                             
/*SDL_ShowCursor (SDL_DISABLE);*/
                                                                                              
  atexit (SDL_Quit);           
  /*  Sound_Init();*/

 }

 if (cap2==1){
   if (cap1==1 && cap3==1) {
if (SDL_Init (SDL_INIT_VIDEO|SDL_INIT_AUDIO) < 0) {
     fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
     exit (1);
}
  Sound_Init();
}
   else {
if (SDL_Init (SDL_INIT_VIDEO) < 0) {
     fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
     exit (1);
   }
   }
   SDL_ShowCursor (SDL_DISABLE);
   atexit (SDL_Quit);

   ch=g_list_nth_data (vm.channels,vm.mirror->vidchan);


(SDLchannel *)sd=ch->data;

   /* video init */

  bpp =                                                                         
    SDL_VideoModeOK (1024, 768, 8, SDL_HWSURFACE | SDL_FULLSCREEN | SDL_HWPALETTE);
                                                                
(SDL_Surface *)sd->screen =                                                                      
    SDL_SetVideoMode (1024, 768, bpp,SDL_HWSURFACE | SDL_FULLSCREEN | SDL_HWPALETTE);
          
                                                                      
                                /* deal with colours */                                                         
  for (i=0;i<256;i++){                                                            
    sd->colors[i].r=0;                                                                  
    sd->colors[i].g=i;                                                                  
    sd->colors[i].b=0;                                                                  
  }                      
  SDL_SetColors((SDL_Surface *)sd->screen, sd->colors, 0, 256);           

 }

 /* deal with filesystem channel */
 
if (cap3==1){
   (struct M_filesystem *)fc=vm.mirror->data3;   
   fc->cap1=cap1;
   fc->cap2=cap2;
}

  signal (SIGTERM, sig_term);
  signal (SIGINT, sig_term);
  signal (SIGSEGV, sig_term);

  while (fnoc==1) {
    Cell *cell;
    pcur = vm.cur_cell;
/*    semop (semid, semtakeover, 2);*/
    if (shared->reboot) {
/*      memcpy (vm.mem, &program, sizeof (program));*/
     /* memset (shared->xp, 0, sizeof (int) * XP_TAIL);*/
fd=open("/dev/urandom", O_RDONLY);

read (fd,vm.mem,vm.memsz);


      shared->pxp = 0;
      resetMachine (&vm);
      shared->reboot = 0;
    }
    step (&vm);
    if (pcur == vm.cur_cell) {
      vm.cur_cell = g_list_next (pcur);
      if (!vm.cur_cell) {
	vm.cur_cell = g_list_first (vm.cells);
cnt++;
if (cnt==WHEN){
  mutate(&vm);
cnt=0;

	gcollect(&vm);}
      }
    }
    cell = ((Cell *) (vm.cur_cell->data));
     shared->xp[shared->pxp++] = (cell->regs.n.xp % cell->len) + cell->offs;
    if (shared->pxp >= XP_TAIL)
      shared->pxp = 0;
    shared->cc_offs = cell->offs;
    shared->cc_len = cell->len;
 /*   semop (semid, semgiveup, 1);
    usleep (opt.delay);*/
    
    /* do sdl key detct bit if needs be */
    /* if (issdl==1)*/ 
 if (SDL_PollEvent (&event)) {                                               
       switch (event.type) {                                                     
             case SDL_KEYDOWN:                                                         
                     switch (event.key.keysym.sym) {      
case 'q':
fnoc=0;
}
   break;
         default:
                 break;
                 

}}

   
  }
  SDL_Quit();
}
